"""Helpers for testing with the kube package.

This implements a :class:`StubAPIServerProxy` which you can program
yourself and does not require a real Kubernetes cluster to work.  The
idea is that it can be used as a drop-in replacement for the real
:class:`kube.APIServerProxy` for tests.

Currently this is not yet exposed as a public API but used in our own
tests.
"""

import binascii
import collections
import copy
import datetime
import hashlib
import os
import random
import time
import types
import urllib.parse
import uuid

import pyrsistent

from kube import _error


class StubResponse:
    """A fake requests.Response class for StubAPIError"""

    def __init__(self, status_code, message):
        self.status_code = status_code
        self._message = message

    def json(self):
        """Return fake JSON data"""
        return {'status_code': self.status_code,
                'message': self._message if self._message else ''}


class StubAPIError(_error.APIError):
    """A kube.APIError subclass raisable without a response."""

    def __init__(self, status, message=None):  # pylint: disable=super-init-not-called
        response = StubResponse(status, message)
        super().__init__(response, message)


class StubAPIServerProxy:
    """An APIServerProxy which emulates a real k8s API server.

    The external API matches :class:`kube.APIServerProxy`, so for
    detailed docs see that class.
    """
    _Path = collections.namedtuple('Path', ['namespace', 'resource', 'name'])

    def __init__(self, base_url='http://localhost:8001/api/v1/', master=None):
        if not base_url.endswith('/'):
            base_url += '/'
        self._base_url = base_url
        if master is not None:
            self._master = master
        else:
            self._master = StubMaster()

    def close(self):
        """Close the underlying connections."""
        pass

    def urljoin(self, *path):
        """Wrapper around urllib.parse.urljoin for the configured base URL."""
        return urllib.parse.urljoin(self._base_url, '/'.join(path))

    def _parse_path(self, path):
        """Parse the path tuple.

        Returns a namedtuple of ``(namespace, resource, name)`` where
        both ``namespace`` and ``name`` can be *None* if they are not
        present in the path.
        """
        url = urllib.parse.urlsplit(self.urljoin(*path))
        items = url.path.split('/')
        base = urllib.parse.urlsplit(self._base_url).path.split('/')[:-1]
        items = items[len(base):]
        if len(items) > 2 and items[0] == 'namespaces':
            namespace = items[1]
            items = items[2:]
        else:
            namespace = None
        resource = items[0]
        try:
            name = items[1]
        except IndexError:
            name = None
        return self._Path(namespace, resource, name)

    def get(self, *path, **params):  # pylint: disable=unused-argument
        """HTTP GET the relative path from the API server."""
        namespace, resource_name, name = self._parse_path(path)
        try:
            resource = getattr(self._master, resource_name)
            if name is not None:
                data = resource.get(name, namespace=namespace)
            else:
                data = resource.select(namespace)
        except (KeyError, AttributeError):
            raise StubAPIError(404, 'No such resource item found')
        else:
            return pyrsistent.freeze(data)

    def post(self, *path, json=None, **params):  # pylint: disable=unused-argument
        """HTTP POST the relative path to the API server."""
        namespace, resource_name, _ = self._parse_path(path)
        try:
            resource = getattr(self._master, resource_name)
        except AttributeError:
            raise StubAPIError(404, 'No such resource item found')
        try:
            name = json['metadata']['name']
        except KeyError:
            basename = json['metadata']['generateName']
            name = basename + binascii.hexlify(os.urandom(3)).decode()
        rev = resource.revisions[-1]
        existing_names = {i['metadata']['name'] for i in rev.items}
        if name in existing_names:
            raise StubAPIError(409, 'Resource with name already exists')
        return pyrsistent.freeze(resource.create(namespace=namespace,
                                                 name=name,
                                                 data=json))

    def delete(self, *path, json=None, **params):  # pylint: disable=unused-argument,unused-variable
        """HTTP DELETE the relative path to the API server."""
        namespace, resource_name, item_name = self._parse_path(path)
        try:
            resource = getattr(self._master, resource_name)
        except AttributeError:
            raise StubAPIError(404, 'No such resource item found')
        try:
            resource.remove(item_name, namespace)
        except LookupError:
            raise StubAPIError(404, 'No such resource item found')

    def patch(self, *path, patch=None):
        """HTTP PATCH as application/strategic-merge-patch+json."""
        ppath = self._parse_path(path)
        try:
            resource = getattr(self._master, ppath.resource)
        except AttributeError:
            raise StubAPIError(404, 'No such resource item found')
        return resource.patch(ppath.name, patch, namespace=ppath.namespace)

    def watch(self, *path, version=None, fields=None):  # pylint: disable=unused-variable
        """Watch a list resource for events.

        Note that this ignores a named item in the resource as that is
        not really supported by k8s anyway.
        """
        path = self._parse_path(path)
        resource = getattr(self._master, path.resource)
        changes_iter = resource.watch(version)
        return StubJSONWatcher(resource, changes_iter)


class StubMaster:
    """A stub Kubernetes master server.

    This contains an API to manipulate the resources like the real k8s
    master API server does.
    """

    NAMESPACE_DATA = {
        'spec': {
            'finalizers': [
                'kubernetes',
            ],
        },
        'status': {
            'phase': 'Active',
        },
    }
    NODE_DATA = {
        'metadata': {           # Some of this is filled in automatically.
            'labels': {
                'kubernetes.io/hostname': 'kube-testnode-abc1',
            },
        },
        'spec': {
            'externalID': '17349516783711110824',
            'podCIDR': '10.120.0.0/24',
            'providerID': 'gce://project/europe-west1-d/kube-testnode-abc1',
        },
        'status': {
            'addresses': [
                {
                    'address': '10.240.0.2',
                    'type': 'InternalIP',
                },
                {
                    'address': '104.155.29.246',
                    'type': 'ExternalIP',
                }
            ],
            'capacity': {
                'cpu': '1',
                'memory': '1745136Ki',
                'pods': '40',
            },
            'conditions': [
                {
                    'lastHeartbeatTime': '2015-11-30T15:45:10Z',
                    'lastTransitionTime': '2015-11-07T20:25:16Z',
                    'reason': 'kubelet is posting ready status',
                    'status': 'True',
                    'type': 'Ready',
                }
            ],
            'nodeInfo': {
                'bootID': 'ecfa1657-c5f1-4388-b132-2bc5516cdb7c',
                'containerRuntimeVersion': 'docker://1.6.2',
                'kernelVersion': '3.16.0-0.bpo.4-amd64',
                'kubeProxyVersion': 'v1.0.6',
                'kubeletVersion': 'v1.0.6',
                'machineID': '',
                'osImage': 'Debian GNU/Linux 7 (wheezy)',
                'systemUUID': 'A3301D3F-154A-F537-9F39-34DF751661ED',
            },
        },
    }

    def __init__(self):
        self.nodes = StubResource('NodeList')
        self.nodes.create(namespace=None,
                          name='kube-testnode-abc1',
                          data=self.NODE_DATA)
        self.namespaces = StubResource('NamespaceList')
        self.namespaces.create(namespace=None,
                               name='default',
                               data=self.NAMESPACE_DATA)
        self.pods = StubResource('PodList', create_cb=self._create_pod)
        self._pod_count = 1
        self.replicationcontrollers = StubResource(
            'ReplicationControllerList', create_cb=self._create_replicaset)
        self.services = StubResource('ServiceList')
        self.secrets = StubResource('SecretList',
                                    create_cb=self._create_secret)

    def _create_pod(self, data):
        """Callback to customise creation of pod resource items."""
        data['status']['phase'] = 'Running'
        start_time = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
        data['status']['startTime'] = start_time
        data['status']['podIP'] = '192.0.2.{}'.format(self._pod_count)
        self._pod_count += 1
        data['status']['hostIP'] = '192.51.100.1'
        base_hash = hashlib.sha256(data['metadata']['namespace'].encode())
        base_hash.update(data['metadata']['name'].encode())
        data['status']['containerStatuses'] = []
        for spec in data['spec']['containers']:
            c_hash = base_hash.copy()
            c_hash.update(spec['name'].encode())
            cid_hash = c_hash.copy()
            cid_hash.update(b'containerID')
            iid_hash = c_hash.copy()
            iid_hash.update(b'imageID')
            status = {
                'containerID': 'docker://{}'.format(cid_hash.hexdigest()),
                'name': spec['name'],
                'ready': True,
                'restartCount': 0,
                'image': spec['image'],
                'imageID': 'docker://{}'.format(iid_hash.hexdigest()),
                'state': {
                    'running': {
                        'startedAt': start_time,
                    },
                },
            }
            data['status']['containerStatuses'].append(status)
        return data

    @staticmethod
    def _create_replicaset(data):
        """Callback to customise creation of ReplicaSet resource items."""
        data['status']['replicas'] = data['spec']['replicas']
        data['status']['observedGeneration'] = 1
        data['status']['fullyLabeledReplicas'] = 1
        return data

    @staticmethod
    def _create_secret(data):
        """Callback to default to the Opaque secret type."""
        if 'type' not in data:
            data['type'] = 'Opaque'
        return data


class StubResource:
    """Represent a stubbed out resource.

    This contains the API which to manage a single k8s resource in a
    generic way.

    :param  kind: The resource kind, e.g. NodeList, PodList, etc.  Note
       that the base resource must always be a *List in k8s.
    :type kind: str
    :param create_cb: An optional callback which will be called when a
       new item is created.  It should take the dict of data and return
       the, optionally modified, same dict of data.
    :type create_cb: types.FunctionType
    """
    Revision = collections.namedtuple('Revision', ['rev', 'items'])
    Revision.__doc__ = """Represents a revision of a k8s resource list.

    Each revision has the integer ``rev`` number and the ``items`` as
    a :class:`pyrsistent.PSet` of :class:`pyrsistent.PMap` instance
    containing their raw data.
    """

    def __init__(self, kind, create_cb=None):
        self.kind = kind
        self._create_cb = create_cb
        self.name = kind[:-4].lower() + 's'
        self.api_version = 'v1'
        initial_rev = random.randint(1, 2000000)
        initial_items = pyrsistent.pset()
        self.revisions = collections.deque(
            [self.Revision(initial_rev, initial_items)], maxlen=200)

    def __repr__(self):
        return '<StubResource {0.kind}>'.format(self)

    def create(self, namespace, name, data):
        """Create a new resource item.

        This adds a new named item to the resource list.  This is
        fairly crude for now and does not match k8s creation semantics
        yet, just a way to get a new resource item.  Note that some
        info in the metadata should not be provided and will be filled
        in automatically:

        - apiVersion
        - kind
        - metadata.creationTimestamp
        - metadata.labels
        - metadata.name
        - metadata.resourceVersion
        - metadata.selfLink
        - metadata.uid

        If you do provide one of these items they will overwrite the
        default ones and you can risk creating an incorrect item,
        e.g. by having a mismatching name.

        :param namespace: The namespace this resource belongs too.
           Use ``"default"`` for the default namespace.  Use ``None``
           for resources which do not belong to namespaces like
           NodeList.
        :type namespace: str or None.
        :param name: The name of the item to create.
        :type name: str
        :param data: The raw data for the item.
        :type data: collections.abc.Mapping
        """
        last = self.revisions[-1]
        names = [i['metadata']['name'] for i in last.items
                 if i['metadata'].get('namespace') == namespace]
        if name in names:
            raise KeyError('Name already exists')
        rev = last.rev + 1
        selflink = '/api/{r.api_version}{ns}/{r.name}/{name}'.format(
            r=self,
            ns='/namespaces/{}'.format(namespace) if namespace else '',
            name=name,
        )
        default_meta = {
            'creationTimestamp':
                datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ'),
            'labels': {},
            'name': name,
            'resourceVersion': str(rev),
            'selfLink': selflink,
            'uid': str(uuid.uuid4()),
        }
        if namespace is not None:
            default_meta['namespace'] = namespace
        new_item = copy.deepcopy(data)
        default_meta.update(new_item.get('metadata', {}))
        new_item['metadata'] = default_meta
        if 'spec' not in new_item:
            new_item['spec'] = {}
        if 'status' not in new_item:
            new_item['status'] = {}
        if 'kind' not in new_item:
            new_item['kind'] = self.kind.rstrip('List')
        if 'apiVersion' not in new_item:
            new_item['apiVersion'] = self.api_version
        if self._create_cb:
            new_item = self._create_cb(new_item)
        new_items = last.items.add(pyrsistent.freeze(new_item))
        self.revisions.append(self.Revision(rev, new_items))
        return new_item

    def get(self, name, namespace=None):
        """Retreive a named item from the resource.

        This will retrieve the latest version of the named resource in
        the given namespace.

        :param str name: The name of the item to retreive.
        :param str namespace: The namespace of the item to retrieve.

        :returns: The mapping describing the item.
        :rtype: pyrsistent.PMap

        :raises KeyError: When the item is not found.
        """
        last = self.revisions[-1]
        items = {i for i in last.items
                 if (i['metadata'].get('namespace') == namespace and
                     i['metadata']['name'] == name)}
        return items.pop()

    def remove(self, name, namespace=None):
        """Remove the named item."""
        last = self.revisions[-1]
        for item in last.items:
            if (item['metadata'].get('namespace') == namespace and
                    item['metadata']['name'] == name):
                break
        else:
            raise LookupError('No such item')
        new_items = last.items.remove(item)  # pylint: disable=undefined-loop-variable
        self.revisions.append(self.Revision(last.rev+1, new_items))

    def patch(self, name, patch, namespace=None):
        """Apply a strategic merge patch to the named item in the resource.

        This creates a new version of the resource with the named item
        modified from the changes specified by the patch.

        :param str name: The name of the resource item.
        :param dict patch: The patch as deserialised JSON.
        :param str namespace: The namespace of the resource item.

        :returns: The new item after the patch is applied.
        :rtype: pyrsistent.PMap
        """
        old = self.get(name, namespace)
        tmp = pyrsistent.thaw(old)
        self._patch(tmp, patch)
        last = self.revisions[-1]
        tmp['metadata']['resourceVersion'] = str(last.rev + 1)
        new = pyrsistent.freeze(tmp)
        items = last.items.remove(old).add(new)
        self.revisions.append(self.Revision(last.rev + 1, items))
        return new

    def _patch(self, obj, patch):
        """Patch a dict object.

        This can recusively apply the logic required to patch a dict.
        It is a JSON Merge Patch (:rfc:`7396`) and not the k8s
        strategic merge patch.

        :param obj: The object to patch.
        :type obj: dict, *must* be mutable.
        :param patch: The patch *for this object*, for a nested object
           this can be a subset of the patch just pertaining this
           object.
        :type patch: collections.abc.Mapping.
        """
        for key, value in patch.items():
            if value is None:
                try:
                    del obj[key]
                except KeyError:
                    pass
            elif isinstance(value, collections.abc.Mapping):
                self._patch(obj[key], patch[key])
            else:
                obj[key] = value

    def select(self, namespace=None):
        """Select multiple items from the resource.

        Currently can only select all or filter on namespace.

        :param str namespace: Restrict selected items to given
           namespace.
        """
        last = self.revisions[-1]
        if namespace is not None:
            items = pyrsistent.pset(
                i for i in last.items
                if i['metadata'].get('namespace') == namespace
            )
        else:
            items = last.items
        return pyrsistent.freeze({
            'apiVersion': self.api_version,
            'items': copy.deepcopy(items),
            'kind': self.kind,
            'metadata': {
                'resourceVersion': str(last.rev),
                'selfLink': '/api/{}/{}'.format(
                    self.api_version,
                    self.kind.rstrip('List').lower() + 's',
                ),
            },
        })

    def _nextrev(self, rev):
        """Return the next revision number for the resource.

        Use ``None`` to get the first revision number.
        """
        for resource in self.revisions:
            if rev is None or resource.rev > rev:
                return resource
        else:                   # pylint: disable=useless-else-on-loop
            raise LookupError

    def watch(self, version=None):
        """Return generator yielding changes for this resource.

        Note that this will raise StopIteration when there are no more
        changes.  To avoid pre-maturely terminating the generator use
        the :meth:`wait` method to find out if more data is available.

        :param int version: The optional version to start watching from.
        """
        if version is not None:
            version = int(version)
            last = self._nextrev(version - 1)
            curr = self._nextrev(version)
        else:
            last = self.revisions[0]
            curr = self.revisions[1]
        while True:
            yield self._compare_revisions(last, curr)
            last = curr
            try:
                curr = self._nextrev(last.rev)
            except LookupError:
                return

    @staticmethod
    def _compare_revisions(prev, curr):
        """Compare two revisions and generate change object.

        This generates the change object for :meth:`watch`.
        """
        if len(prev.items) < len(curr.items):
            new = curr.items - prev.items
            assert len(new) == 1
            return pyrsistent.pmap({
                'type': 'ADDED',
                'object': next(iter(new)),
            })
        elif len(prev.items) > len(curr.items):
            rem = prev.items - curr.items
            assert len(rem) == 1
            return pyrsistent.pmap({
                'type': 'DELETED',
                'object': next(iter(rem)),
            })
        else:
            prevd = {i['metadata']['uid']: i for i in prev.items}
            currd = {i['metadata']['uid']: i for i in curr.items}
            prev_uids = sorted(prevd.keys())
            curr_uids = sorted(currd.keys())
            assert prev_uids == curr_uids
            for uid, prev_item in prevd.items():
                if prev_item != currd[uid]:
                    return pyrsistent.pmap({
                        'type': 'MODIFIED',
                        'object': currd[uid],
                    })

    def wait(self, last_rev, timeout):
        """Wait for the next revision to be ready.

        :param int last_rev: The previous revision.  If ``None`` this
           will wait for the first revision.
        :param int timeout: The time to wait in seconds.

        :returns: ``True`` if the next revision is available,
           ``False`` if not.
        :rtype: bool
        """
        deadline = time.monotonic() + timeout
        while time.monotonic() <= deadline:
            try:
                self._nextrev(last_rev)
            except LookupError:
                time.sleep(0.02)
            else:
                return True
        return False


class StubJSONWatcher:
    """A stub JSONWatcher instance to work with :class:`StubMaster`.

    This matches the :class:`kube._watch.JSONWatcher` class but
    works with the :class:`StubMaster` instead.  See the original
    class for full docstrings.

    Note that the constructor has a different signature as this is not
    part of the public API.  Also note that it does provide a (fake)
    ``._sock.fileno()`` method which will return -1 after
    :meth:`close` has been called, this is to support the tests.

    :param resource: The resource this instance is watching.
    :type resource: StubResource
    :param changes_iter: The generator from :meth:`StubResource.wait`.
    :type changes_iter: generator
    """

    def __init__(self, resource, changes_iter):
        self._resource = resource
        self._changes_iter = changes_iter
        self._last_rev = None
        self._sock = types.SimpleNamespace(fileno=self._stub_fileno)
        self._fake_fileno = 42

    def _stub_fileno(self):
        """Fake socket.fileno() function."""
        return self._fake_fileno

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self, *, timeout=None):
        """Return the next line."""
        if timeout is not None:
            ready = self._resource.wait(self._last_rev, timeout)
            if not ready:
                raise TimeoutError('No new revision available')
        item = next(self._changes_iter)
        return pyrsistent.thaw(item)

    def skip(self, version):
        """Skip past given version if possible."""
        pass

    def close(self):
        """Close the iterator."""
        self._fake_fileno = -1

    def __enter__(self):
        pass

    def __exit__(self, exc_val, exc_type, traceback):
        self.close()



"""Interface for Service resources."""

import ipaddress

from kube import _base
from kube import _error


class ServiceView(_base.BaseView, _base.ViewABC):
    """View of the Service resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.ServiceList
    resource = 'services'


class ServiceItem(_base.BaseItem, _base.ItemABC):
    """A Service in the Kubernetes cluster.

    :param cluster: The cluster this Service exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.Service
    resource = 'services'

    @property
    def loadbalancer_ingress(self):
        """The loadbalancer ingress endpoints.

        This is a set of ingress endpoints in use by the
        loadbalancer.  Depending on the infrastructure the cluster
        runs on the endpoint can be either an
        :class:`ipaddress.IPv4Address`, :class:`ipaddress.IPv6Address`
        or a hostname as a string.
        """
        try:
            ingress = self.raw['status']['loadBalancer']['ingress']
        except KeyError:
            raise _error.StatusError
        endpoints = set()
        for endpoint in ingress:
            if 'ip' in endpoint:
                endpoints.add(ipaddress.ip_address(endpoint['ip']))
            elif 'hostname' in endpoint:
                endpoints.add(endpoint['hostname'])
            else:
                raise _error.StatusError
        return endpoints



"""Helper functions for WATCH."""

import collections
import enum
import io
import json
import select
import socket
import time
import urllib.parse

import http_parser.parser

from kube import _error


class JSONWatcher:
    """Iterator to help with WATCH on Kubernetes resources.

    This iterator will issue a WATCH command to the given path and
    yield JSON objects as they become available.  The standard
    iterator interface is blocking while the extended :meth:`next`
    interface allows non-blocking use.

    Note that you should only want a resource and not any of it's
    items inside it.  Resources always have a ``*List`` kind.

    :param proxy: The proxy this watcher is for.
    :type proxy: kube.APIServerProxy
    :param path: The path of the resource to watch, see
       :meth:`kube.APIServerProxy.urljoin` for details.
    :type path: str
    :param version: The resourceVersion from which to start watching
       the resource.
    :type version: str
    :param fields: A dict of fields which must match their values.
       This is a limited form of the full fieldSelector format, it is
       limited because filtering is done at client side for
       consistecy.
    :type fields: dict

    :raises kube.APIError: When the API server does not respond
       correctly.

    :ivar cluster: The :class:`kube.Cluster` instance.
    """

    def __init__(self, proxy, *path, version=None, fields=None):
        params = {'watch': 'true'}
        if version:
            params['resourceVersion'] = version
        self._fields = fields if fields else {}
        query = '&'.join(k + '=' + v for k, v in params.items())
        url = urllib.parse.urlsplit(proxy.urljoin(*path) + '?' + query)
        self._buffer = collections.deque()
        self._pending_data = b''
        self._sock = self._connect(url)
        self._parser = self._make_request(self._sock, url)
        self._fill_buffer(timeout=0)

    @staticmethod
    def _connect(url):
        """Connect to the URL and perform the initial request.

        This will create a socket and connect it to the server for the
        URL.

        :param url: The URL to connect too.
        :type url: urllib.parse.SplitResult, the result of
           urllib.parse.spliturl().

        :returns: The connected socket.
        :raises APIError: When failing to connect to the server.
        """
        addrinfo = socket.getaddrinfo(url.hostname, url.port or 80,
                                      type=socket.SOCK_STREAM)
        for family, kind, proto, _, sockaddr in addrinfo:
            sock = socket.socket(family, kind, proto)
            try:
                sock.connect(sockaddr)
            except ConnectionRefusedError:
                sock.close()
                continue
            else:
                return sock
        else:                   # pylint: disable=useless-else-on-loop
            raise _error.APIError('Failed to connect to {}'
                                  .format(urllib.parse.urlunsplit(url)))

    @staticmethod
    def _make_request(sock, url):
        """Make the GET/watch request.

        This performs the initial GET request to start the watch
        operation and reads the headers from the response.  Some of
        the body may already have been read as well and will be
        available in the returned http parser instance.

        :param sock: The connected socket.
        :type sock: socket.socket
        :param url: The split URL from urllib.parse.urlsplit().
        :type url: urllib.parse.SplitResult

        :returns: The HTTP parser instance.
        :rtype: http_parser.parser.HttpParser

        :raises APIError: For errors from the k8s API server.
        """
        request = ('GET {0.path}?{0.query} HTTP/1.1\r\n'
                   'Host: {0.hostname}\r\n'
                   '\r\n'.format(url))
        class DummyReq:
            """Request for APIError purposes."""
            body = request
            status_code = 0
        parser = http_parser.parser.HttpParser()  # pylint: disable=no-member
        sock.send(request.encode('ASCII'))
        while not parser.is_headers_complete():
            data = sock.recv(io.DEFAULT_BUFFER_SIZE)
            if not data:
                raise _error.APIError(DummyReq(), 'Did not receive response')
            nreceived = len(data)
            nparsed = parser.execute(data, nreceived)
            DummyReq.status_code = parser.get_status_code()
            if nparsed != nreceived:
                raise _error.APIError(DummyReq(), 'HTTP parsing failed')
        if not parser.is_chunked():
            raise _error.APIError(DummyReq(), 'Not a chunked HTTP response')
        return parser

    def __iter__(self):
        return self

    def __next__(self):
        """Return the next :class:`kube.WatchEvent` instance.

        This is the equivalent of :meth:`next` with `timeout=None` so
        will block if no more events are queued.
        """
        return self.next()

    def next(self, *, timeout=None):
        """Return the next line.

        :param timeout: If specified a timeout in seconds.  If no new
           event occurred in this time a :class:`TimeoutError` is
           raised.  Set to ``0`` for a non-blocking call.
        :type timeout: int or float.

        :raises TimeoutError: In case retrieving the next event
           timed out.
        """
        if self._buffer:
            return self._buffer.popleft()
        end = time.monotonic() + timeout
        remaining = timeout
        while remaining >= 0:
            self._fill_buffer(timeout=remaining)
            if self._buffer:
                return self._buffer.popleft()
            elif self._sock.fileno() == -1:
                raise StopIteration
            remaining = end - time.monotonic()
        raise TimeoutError

    def skip(self, version):
        """Skip ahead past the given version on the resource.

        This iterates over already buffered events and compares the
        version number of each resource item with the given version.
        If an item with the given version is found, all items up to
        and including the item are dropped from the buffer, otherwise
        the buffer remains unmodified.

        :param version: The resourceVersion of the item you want to
           skip past.
        :type version: bytes

        :raises TypeError: If the objects do not have a valid
           resourceVersion.
        """
        new_buffer = []
        keep = False
        for obj in self._buffer:
            if keep:
                new_buffer.append(obj)
                continue
            try:
                item_version = obj['object']['metadata']['resourceVersion']
            except (KeyError, TypeError) as err:
                raise TypeError('No resourceVersion in object') from err
            else:
                if item_version == version:
                    keep = True
        if new_buffer:
            self._buffer = collections.deque(new_buffer)

    def _fill_buffer(self, *, timeout=None):
        """Fill the buffer with fully received lines.

        This fills :attr:`_buffer` with fully received lines,
        i.e. lines ending in ``\n``.  New data is read regardless of
        whether the buffer already contains lines.  Any partially
        received lines are cached in :attr:`_pending_data` for the
        next call to this method.
        """
        # bytes.split() will return an empty bytestring if the line is
        # already terminated so the last item is always pending data.
        data = self._read_data(timeout=timeout)
        data = self._pending_data + data
        lines = data.split(b'\n')
        self._pending_data = lines.pop(-1)
        for line in lines:
            obj = json.loads(line.decode('utf-8'))
            if not self._match_fields(obj['object']):
                continue
            self._buffer.append(obj)

    def _match_fields(self, obj):
        """Check whether the object matches the fields configured."""
        results = []
        for name, value in self._fields.items():
            item = obj
            for key in name.split('.'):
                try:
                    item = item[key]
                except KeyError:
                    return False
            results.append(item == value)
        return all(results)

    def _read_data(self, *, timeout=None):
        """Read a chunk of data.

        This will try to read all data available.  If no data is
        available it will wait up to *timeout* seconds for data.  Once
        any data is received all available data is returned without
        further delays.

        :param int timeout: The maxium time to wait for data, or wait
           forever if None.

        :returns: The received data or an empty string if no data was
           received.
        :rtype: bytes
        :raises: StopIteration if the socket was closed.
        """
        if self._sock.fileno() == -1:
            return ''
        sock = self._sock
        readers = [sock]
        writers = []
        while True:
            rlist, _, _ = select.select(readers, writers, writers, timeout)
            if not rlist:
                break
            timeout = 0
            chunk = sock.recv(io.DEFAULT_BUFFER_SIZE)
            if not chunk:
                sock.close()
                break
            nreceived = len(chunk)
            nparsed = self._parser.execute(chunk, nreceived)
            if nparsed != nreceived:
                raise _error.APIError('Failed to parse HTTP body')
        return self._parser.recv_body()

    def close(self):
        """Close the iterator.

        This releases the socket.  Once the already buffered data is
        consumed :class:`StopIteration` will be raised.
        """
        self._sock.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_val, exc_type, traceback):
        self.close()


class WatchEventType(enum.Enum):
    """Represents the type of watch event."""
    MODIFIED = 'MODIFIED'
    ADDED = 'ADDED'
    DELETED = 'DELETED'
    ERROR = 'ERROR'


WatchEvent = collections.namedtuple('WatchEvent', ('evtype', 'item'))
WatchEvent.MODIFIED = WatchEventType.MODIFIED
WatchEvent.ADDED = WatchEventType.ADDED
WatchEvent.DELETED = WatchEventType.DELETED
WatchEvent.ERROR = WatchEventType.ERROR
WatchEvent.__doc__ = """\
Represents an event and it's object generated from watching a resource.

:cvar MODIFIED: A convenience alias of :attr:`WatchEventType.MODIFIED`.

:ivar evtype: The event type enumeration.
:type evtype: :class:`WatchEventType`
:ivar item: The object which is the subject of the event.
:type item: This depends on who created this :class:`WatchEvent`.
"""


class ResourceWatcher:
    """Watcher for a resource.

    This is an iterator yielding watch events in either a blocking or
    non-blocking way, for non-blocking use ``.next(timeout=0)``.  It
    uses a :class:`JSONWatcher` instance for retrieving the actual
    events, which must be configured correctly to return events for
    the same resource as this watcher is for.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster

    :param jsonwatcher: A correctly configured watcher instance which
       yields the decoded JSON objects.
    :type jsonwatcher: JSONWatcher

    :param itemcls: A constructor for the resource item being watched.
    :type itemcls: A callable, usually a class.
    """

    def __init__(self, cluster, jsonwatcher, itemcls):
        self.cluster = cluster
        self._watcher = jsonwatcher
        self._itemcls = itemcls

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self, *, timeout=None):
        """Return the next watch event.

        :param timeout: The maximum time to wait for a new event.  Not
           specifying this will block forever until a new event
           arrives, otherwise a :class:`TimeoutError` is raised if no
           new event was received in time.
        :type timeout: int or float

        :raises TimeoutError: When no new event is available after the
           specified timeout.
        """
        obj = self._watcher.next(timeout=timeout)
        return WatchEvent(WatchEventType(obj['type']),
                          self._itemcls(self.cluster, obj['object']))

    def close(self):
        """Close the iterator and release it's resources.

        This releases the underlying socket.
        """
        self._watcher.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_val, exc_type, traceback):
        self.close()



"""Tools to work with a Kubernetes cluster.

This module contains the toplevel tools to work with a Kubernets
cluster and it's API server.
"""

import urllib.parse

import requests

from kube import _base
from kube import _error
from kube import _namespace
from kube import _node
from kube import _pod
from kube import _replicaset
from kube import _secret
from kube import _service
from kube import _util
from kube import _watch


class Cluster:
    """A Kubernetes cluster.

    The entrypoint to control a Kubernetes cluster.  There is only one
    connection mechanism, which is via a local API server proxy.  This
    is normally achieved by running ``kubectl proxy``.

    :param str url: The URL of the API server.
    """
    _CLS_MAP = {
        _base.Kind.Node: _node.NodeItem,
        _base.Kind.NodeList: _node.NodeView,
        _base.Kind.Namespace: _namespace.NamespaceItem,
        _base.Kind.NamespaceList: _namespace.NamespaceView,
        _base.Kind.ReplicaSet: _replicaset.ReplicaSetItem,
        _base.Kind.ReplicaSetList: _replicaset.ReplicaSetView,
        _base.Kind.Pod: _pod.PodItem,
        _base.Kind.PodList: _pod.PodView,
        _base.Kind.Service: _service.ServiceItem,
        _base.Kind.ServiceList: _service.ServiceView,
        _base.Kind.Secret: _secret.SecretItem,
        _base.Kind.SecretList: _secret.SecretView,
    }

    def __init__(self, url='http://localhost:8001/api/'):
        if not url.endswith('/'):
            url += '/'
        api_url = urllib.parse.urljoin(url, 'v1/')
        self.proxy = APIServerProxy(api_url)
        self.nodes = self.kindimpl(_base.Kind.NodeList)(self)
        self.namespaces = self.kindimpl(_base.Kind.NamespaceList)(self)
        self.replicasets = self.kindimpl(_base.Kind.ReplicaSetList)(self)
        self.pods = self.kindimpl(_base.Kind.PodList)(self)
        self.services = self.kindimpl(_base.Kind.ServiceList)(self)
        self.secrets = self.kindimpl(_base.Kind.SecretList)(self)

    def close(self):
        """Close and clean up underlying resources."""
        self.proxy.close()

    @classmethod
    def kindimpl(cls, kind):
        """Return the class which implements the resource kind.

        :param kind: The :class:`kube.Kind` instance.
        :type kind: kube.Kind

        :returns: A class implementing either :class:`kube.ViewABC` or
           :class:`kube.ItemABC` depending on the kind.

        :raises ValueError: If the kind is not know.
        """
        try:
            return cls._CLS_MAP[kind]
        except KeyError as err:
            raise ValueError('Unknown kind') from err

    def create(self, data, namespace=None):
        """Create a new resource item.

        :param data: The specification to create the resource from,
           this must include the ``apiVersion``, ``kind``,
           ``metadata`` and ``spec`` fields.  It is usually simply the
           de-serialised YAML but allows you to insert template
           processing if you require so.
        :type data: dict
        :param namespace: Create the resource item in the given
           namespace.  If the ``spec`` includes a namespace this
           namespace must match or an exception will be raised.
        :type namespace: str

        :returns: The newly created item.
        :rtype: A :class:`kube.ViewABC` instance of the right type
           according to the kind of resource item created based on the
           data in the spec.

        :raises kube.APIError: For errors from the k8s API server.
        :raises kube.KubeError: If the spec is incomplete or the kind
           is unknown.
        """
        try:
            kind = _base.Kind(data['kind'])
        except KeyError as err:
            raise _error.KubeError('No resource kind found') from err
        except ValueError as err:
            raise _error.KubeError('Unknown kind') from err
        item_cls = self.kindimpl(kind)
        path = [item_cls.resource]
        if namespace is not None:
            path = ['namespaces', namespace] + path
        new = self.proxy.post(*path, json=data)
        return item_cls(self, new)

    def __enter__(self):
        return self

    def __exit__(self, exc_val, exc_type, traceback):
        self.close()


class APIServerProxy:
    """Helper class to directly communicate with the API server.

    Since most classes need to communicate with the Kubernetes
    cluster's API server in a common way this class helps take care of
    the common logic.  It also keeps the requests session alive to
    enable connection pooling to the API server.

    :param str base_url: The URL of the API, including the API version.
    """

    def __init__(self, base_url='http://localhost:8001/api/v1/'):
        if not base_url.endswith('/'):
            base_url += '/'
        self._base_url = base_url
        self._session = requests.Session()

    def close(self):
        """Close underlying connections.

        Once the proxy has been closed then the it can no longer be used
        to issue further requests.
        """
        self._session.close()

    def urljoin(self, *path):
        """Wrapper around urllib.parse.urljoin for the configured base URL.

        :param path: Individual relative path components, they will be
           joined using "/".  None of the path components should
           include a "/" separator themselves.
        """
        return urllib.parse.urljoin(self._base_url, '/'.join(path))

    def get(self, *path, **params):
        """HTTP GET the relative path from the API server.

        :param str path: Individual relative path components, they
           will be joined using "/".  None of the path components
           should include a "/" separator themselves.
        :param dict params: Extra query parameters for the URL of the
           GET request as a dictionary of strings.

        :returns: The decoded JSON data.
        :rtype: pyrsistent.PMap

        :raises kube.APIError: If the response status is not 200 OK.
        """
        url = self.urljoin(*path)
        response = self._session.get(url, params=params)
        if response.status_code != 200:
            raise _error.APIError(response, 'Failed to GET {}'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def post(self, *path, json=None, **params):
        """HTTP POST to the relative path on the API server.

        :param path: Individual relative path components, they will be
           joined using :meth:`urljoin`.
        :type path: str
        :param json: The body to post, which will be JSON-encoded
           before posting.
        :type json: collections.abc.Mapping
        :param params: Extra query parameters for the URL of the POST
           request.
        :type params: str

        :returns: The decoded JSON data.
        :rtype: pyrsistent.PMap

        :raises kube.APIError: If the response status is not 201
           Created.
        """
        url = self.urljoin(*path)
        response = self._session.post(url, json=json, params=params)
        if response.status_code != 201:
            raise _error.APIError(response, 'Failed to POST {}'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def delete(self, *path, json=None, **params):
        """HTTP DELETE to the relative path on the API server.

        :param path: Individual relative path components, they will be
           joined using :meth:`urljoin`.
        :type path: str
        :param json: The body, which will be JSON-encoded before
           posting.
        :type json: collections.abc.Mapping
        :param params: Extra query parameters for the URL of the
           DELETE request.
        :type params: str

        :returns: The decoded JSON data.
        :rtype: pyrsistent.PMap

        :raises kube.APIError: If the response status is not 200 OK.
        """
        url = self.urljoin(*path)
        response = self._session.delete(url, json=json, params=params)
        if response.status_code != 200:
            raise _error.APIError(response, 'Failed to DELETE {}'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def patch(self, *path, patch=None):
        """HTTP PATCH as application/strategic-merge-patch+json.

        This allows using the Strategic Merge Patch to patch a
        resource on the Kubernetes API server.

        :param str path: Individual relative path components, they
           will be joined using "/".  None of the path components
           should include a "/" separator themselves, unless you only
           provide one component which will be joined to the base URL
           using :func:`urllib.parse.urljoin`.  This case can be
           useful to use the links provided by the API itself
           directly, e.g. from a resource's ``metadata.selfLink``
           field.
        :param dict patch: The decoded JSON object with the patch
           data.

        :returns: The decoded JSON object of the resource after
           applying the patch.

        :raises APIError: If the response status is not 200 OK.
        """
        url = self.urljoin(*path)
        headers = {'Content-Type': 'application/strategic-merge-patch+json'}
        response = self._session.patch(url, headers=headers, json=patch)
        if response.status_code != 200:
            raise _error.APIError(response, 'Failed to PATCH {}:'.format(url))
        else:
            return response.json(cls=_util.ImmutableJSONDecoder)

    def watch(self, *path, version=None, fields=None):
        """Watch a list resource for events.

        This issues a request to the API with the ``watch`` query
        string parameter set to ``true`` which returns a chunked
        response.  An iterator is returned which continuously reads
        from the response, yielding received lines as bytes.

        :param path: The URL path to the resource to watch. See
            :meth:`urljoin`.
        :param str version: The resource version to start watching from.
        :param dict fields: A dict of fields which must match their
           values.  This is a limited form of the full fieldSelector
           format, it is limited because filtering is done at client
           side for consistecy.

        :returns: An special iterator which allows non-blocking
           iterating using a ``.next(timeout)`` method.  Using it as a
           normal iterator will result in blocking behaviour.
        :rtype: :class:`kube._watch.JSONWatcher`.

        :raises APIError: If there is a problem with the API server.
        """
        return _watch.JSONWatcher(self, *path, version=version, fields=fields)



"""Interface for node resources."""

import collections
import enum
import ipaddress

from kube import _base
from kube import _error
from kube import _util


class NodeView(_base.BaseView, _base.ViewABC):
    """View of all the Node resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.  This is here for the :class:`kube.ViewABC`
       compatibility but can not be used for the NodeList resource.  A
       :class:`kube.NamespaceError` is raised when this is not
       ``None``.
    :type namespace: NoneType

    :raises kube.NamespaceError: If instantiated using a namespace.

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.

    :ivar cluster: The :class:`kube.Cluster` instance.
    """
    kind = _base.Kind.NodeList
    resource = 'nodes'

    def __init__(self, cluster, namespace=None):
        if namespace is not None:
            raise _error.NamespaceError('NodeView does not support namespaces')
        super().__init__(cluster, None)

    def fetch(self, name, namespace=None):
        """Retrieve an individual node by name.

        This returns the current verison of the resource item.

        :param str name: The name of the node to retrieve.
        :param str namespace: Must be *None* or a
           :class:`kube.NamespaceError` is raised.  Here only for
           compatibility with the ABC.

        :return: A single :class:`kube.NodeItem` instance.
        :raises LookupError: If the node does not exist.
        :raises kube.APIError: For errors from the k8s API server.
        :raises kube.NamespaceError: If a namespace is used.
        """
        if namespace is not None:
            raise _error.NamespaceError('NodeView does not support namespaces')
        data = _util.fetch_resource(self.cluster, 'nodes', name)
        return NodeItem(self.cluster, data)


class AddressType(enum.Enum):
    """Enumeration of the address types."""
    ExternalIP = 'ExternalIP'
    InternalIP = 'InternalIP'
    Hostname = 'Hostname'


class NodeItem(_base.BaseItem, _base.ItemABC):
    """A node in the Kubernetes cluster.

    See http://kubernetes.io/docs/admin/node/ for details.

    :param cluster: The cluster this node belongs to.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.Node
    resource = 'nodes'
    _Address = collections.namedtuple('Address', ['type', 'addr'])

    @property
    def addresses(self):
        """An iterator of the addresses for this node.

        Each address is a namedtuple with ``(type, addr)`` as fields.
        Known types are in the :class:`kube.AddressType` enumeration.

        An empty list is returned if there are not yet any addresses
        associated with the node.
        """
        status = self.raw.get('status', {})
        for raw in status.get('addresses', []):
            type_ = AddressType(raw['type'])
            addr = ipaddress.ip_address(raw['address'])
            yield self._Address(type_, addr)

    @property
    def capacity(self):
        """The capacity of the node.

        CPU is expressed in cores and can use fractions of cores,
        while memory is expressed in bytes.
        """
        # See http://kubernetes.io/v1.1/docs/design/resources.html for
        # details on resources usage.  This needs to deal with custom
        # resources as well.  The current stub implementation does not
        # match well.
        raise NotImplementedError

    @property
    def conditions(self):
        """List of conditions."""
        raise NotImplementedError



"""Some base classes and ABCs for kube.

While all implementations are supposed to at least register their
classes, subclassing is not really a aim.  With the exception of
:meth:`__repr__` most properties and methods are marked as abstract
and do not have a common implementation even if they could.  To share
code instead helper functions in ``kube._util`` should be used.
"""

import abc
import enum

import pyrsistent

from kube import _error
from kube import _meta
from kube import _util
from kube import _watch


class ViewABC(metaclass=abc.ABCMeta):         # pragma: nocover
    """Represents a view to a collection of resources.

    All top-level resources in Kubernetes have a collection, resources
    of a ``*List`` kind, with some common functionality.  This ABC
    defines views to provide access to resources in collections in a
    uniform way.  Note that a view is not the same as the collection
    resource, e.g. collections resources have some metadata associated
    with them and exist at a particular point in time, they have a
    metadata.resourceVersion, which views do not have.

    It is always possible to create an instance of this without
    needing to do any requests to the real Kubernetes cluster.

    :param cluster: The cluster this resource list is part of.
    :type cluster: kube.Cluster
    :param namespace: The optional namespace this resource list is
       part of.  If the resource list is not part of a namespace this
       will be ``None`` which means it will be a view to all resources
       of a certain type, regardelss of their namespace.
    :type namespace: str

    :raises kube.NamespaceError: When a namespace is provided but the
       resource does not support one.
    """

    @abc.abstractmethod
    def __init__(self, cluster, namespace=None):
        pass

    @property
    @abc.abstractmethod
    def cluster(self):
        """The :class:`kube.Cluster` instance this resource is bound to."""

    @property
    @abc.abstractmethod
    def namespace(self):
        """The optional namespace this view is bound to.

        If the view is not bound to a namespace this will be ``None``,
        including for resources which do not support namespaces.
        """

    @abc.abstractmethod
    def __iter__(self):
        """Iterator of all resources in this collection.

        :raises kube.APIError: for errors from the k8s API server.
        """

    @property
    @abc.abstractmethod
    def kind(self):
        """The kind of the underlying Kubernetes resource.

        This is a :class:`kube.Kind` enum.

        This should be implemented as a static attribute since it
        needs to be available on the class as well as on the instance.
        """

    @property
    @abc.abstractmethod
    def resource(self):
        """The name of the Kubernetes API resource.

        The resource name is used in the construction of the API
        endpoint, e.g. for the API endpoint
        ``/namespaces/default/pods/`` the resource name is ``pods``.
        The resource name is identical for both the resource as well
        as the resource item, e.g. both objects with ``PodList`` and
        ``Pod`` as kind will have a resource name of ``pods``.

        This should be implemented as a static attribute since it
        needs to be available on the class as well as on the instance.
        """

    @abc.abstractmethod
    def fetch(self, name, namespace=None):
        """Retrieve the current version of a single resource item by name.

        If the view itself is associated with a namespace,
        ``self.namespace is not None``, then this will default to
        fetching the resource item from this namespace.  If the view
        is not associated with a namespace, ``self.namespace is
        None``, and the resource requires a namespace then a
        :class:`kube.NamespaceError` is raised.  Note that the
        ``default`` namespace is not automatically used in this case.

        :param str name: The name of the resource.
        :param str namespace: The namespace to fetch the resource
           from.

        :returns: A single instance representing the resource.

        :raises LookupError: If the resource does not exist.
        :raises kube.NamespaceError: For an invalid namespace, either
           because the namespace is required for this resource but not
           provided or the resource does not support namespaces and
           one was provided.
        :raises kube.APIError: For errors from the k8s API server.
        """

    @abc.abstractmethod
    def filter(self, *, labels=None, fields=None):
        """Return an iterable of a subset of the resources.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: An iterator of :class:`kube.ItemABC` instances of
           the correct type for the resrouce which match the given
           selector.

        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: For errors from the k8s API server.
        """

    @abc.abstractmethod
    def watch(self):
        """Watch for changes to any of the resources in the view.

        Whenever one of the resources in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances.

        :raises kube.NamespaceError: Whe the namespace no longer
           exists.
        :raises kube.APIError: For errors from the k8s API server.
        """


class ItemABC(metaclass=abc.ABCMeta):  # pragma: nocover
    """Representation for a kubernetes resource.

    This is the interface all resource items must implement.

    :param cluster: The cluster this resource is bound to.
    :type cluster: kube.Cluster
    :param raw: The decoded JSON representing the resource.
    :type raw: dict
    """

    @abc.abstractmethod
    def __init__(self, cluster, raw):
        pass

    @property
    @abc.abstractmethod
    def cluster(self):
        """The :class:`kube.Cluster` instance this resource is bound to."""

    @property
    @abc.abstractmethod
    def raw(self):
        """The raw decoded JSON representing the resource.

        This behaves like a dict but is actually an immutable view of
        the dict.
        """

    @abc.abstractmethod
    def fetch(self):
        """Fetch the current version of the resource item.

        This will return a new instance of the current resource item
        at it's latest version.  This is useful to see any changes
        made to the object since it was last retrieved.

        :returns: An instance of the relevant :class:`ItemABC`
           subclass.

        :raises kube.APIError: For errors from the k8s API server.
        """
        data = self.cluster.proxy.get(self.meta.link)
        return self.__class__(self.cluster, data)

    @property
    @abc.abstractmethod
    def kind(self):
        """The Kubernetes resource kind of the resource.

        This is a :class:`kube.Kind` enum.

        This should be implemented as a static attribute since it
        needs to be available on the class as well as on the instance.
        """

    @property
    @abc.abstractmethod
    def resource(self):
        """The name of the Kubernetes API resource.

        The resource name is used in the construction of the API
        endpoint, e.g. for the API endpoint
        ``/namespaces/default/pods/`` the resource name is ``pods``.
        The resource name is identical for both the resource as well
        as the resource item, e.g. both objects with ``PodList`` and
        ``Pod`` as kind will have a resource name of ``pods``.

        This should be implemented as a static attribute since it
        needs to be available on the class as well as on the instance.
        """

    @property
    @abc.abstractmethod
    def meta(self):
        """The resource's metadata as a :class:`kube.ObjectMeta` instance."""

    @abc.abstractmethod
    def spec(self):
        """The spec of this node's resource.

        This returns a copy of the raw, decodeed JSON data
        representing the spec of this resource which can be used to
        re-create the resource.
        """

    @abc.abstractmethod
    def watch(self):
        """Watch the resource item for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :returns: An iterator of :class:`kube.WatchEvents` instances
           for the resource item.

        :raises kube.APIError: For errors from the k8s API server.
        """

    @abc.abstractmethod
    def delete(self):
        """Delete the resource item.

        :rtype: None

        :raises APIError: For errors from the k8s API server.
        """


class BaseView:
    """Optional base class to implement ViewABC.

    This base class provides a common implementation for some of the
    :class:`ViewABC` interface abstract mehtods.  Driving from this
    will not automatically make you an :class:`ItemABC` instance, you
    still need to either subclass or register the :class:`ItemABC` as
    well.

    Note the base-class is for resources which reside in a namespace.
    When using this as a base-class for resources which do not have a
    namespace some of the methods can not be used and will need to be
    overwritten.
    """
    # Subclasses must override these:
    kind = None
    resource = None

    def __init__(self, cluster, namespace=None):
        self._cluster = cluster
        self._namespace = namespace

    @property
    def cluster(self):
        """The cluster instance this view is bound to."""
        return self._cluster

    @property
    def namespace(self):
        """The optional namespace this view is bound to."""
        return self._namespace

    def _create_item(self, data):
        """Create an instance of the item.

        :param data: The raw data of the item.
        :type data: dict.
        """
        itemkind = Kind(self.kind.value[:-4])
        itemcls = self.cluster.kindimpl(itemkind)
        return itemcls(self.cluster, data)

    def __iter__(self):
        """Iterator over all items in this view.

        :raise kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            path = ['namespaces', self.namespace, self.resource]
        else:
            path = [self.resource]
        data = self.cluster.proxy.get(*path)
        for item in data['items']:
            yield self._create_item(item)

    def fetch(self, name, namespace=None):
        """Retrieve an individual resource item by name.

        This returns the current verison of the resource item.

        :param str name: The name of the item to retrieve.
        :param str namespace: The namespace of the item to retrieve.

        :return: A single :class:`kube.NodeItem` instance.
        :raises LookupError: If the node does not exist.
        :raises kube.APIError: For errors from the k8s API server.
        :raises kube.NamespaceError: If a namespace is used.
        """
        # Note: This method can not be used by resources which do not
        #       have a namespace as this implementation requires a
        #       namespace.
        ns = namespace or self.namespace
        if not ns:
            raise _error.NamespaceError('Missing namespace')
        data = _util.fetch_resource(self.cluster,
                                    self.resource, name, namespace=ns)
        return self._create_item(data)

    def filter(self, *, labels=None, fields=None):
        """Return an iterable of a subset of the resources.

        :param labels: A label selector expression.  This can either
           be a dictionary with labels which must match exactly, or a
           string label expression as understood by k8s itself.
        :type labels: dict or str
        :param fields: A field selector expression.  This can either
           be a dictionary with fields which must match exactly, or a
           string field selector as understood by k8s itself.
        :type fields: dict or str

        :returns: An iterator of :class:`kube.ItemABC` instances
           of the correct type for the resrouce.

        :raises ValueError: If an empty selector is used.  An empty
           selector is almost certainly not what you want.  Kubernetes
           treats an **empty** selector as *all* items and treats a
           **null** selector as *no* items.
        :raises kube.APIError: for errors from the k8s API server.
        """
        data_iter = _util.filter_list(self.cluster, self.resource,
                                      labels=labels, fields=fields,
                                      namespace=self.namespace)
        for item in data_iter:
            yield self._create_item(item)

    def watch(self):
        """Watch the resource of this view for changes to items.

        Whenever one of the resources items in the view changes a new
        :class:`kube.WatchEvent` instance is yielded.  You can
        currently not control from "when" resources are being watched,
        other then from "now".  So be aware of any race conditions
        with watching.

        :returns: An iterator of :class:`kube.WatchEvent` instances
           where the :attr:`kube.WatchEvent.item` attribute will be of
           the relevant :class:`kube.ItemABC` instance for the kind.

        :raises kube.APIError: For errors from the k8s API server.
        """
        if self.namespace:
            path = ['namespaces', self.namespace, self.resource]
        else:
            path = [self.resource]
        itemlist = self.cluster.proxy.get(*path)
        version = itemlist['metadata']['resourceVersion']
        jsonwatcher = self.cluster.proxy.watch(*path, version=version)
        itemkind = Kind(self.kind.value[:-4])
        itemcls = self.cluster.kindimpl(itemkind)
        return _watch.ResourceWatcher(self.cluster, jsonwatcher, itemcls)

    def __repr__(self):
        if self.namespace is None:
            return '<{0.__class__.__name__}>'.format(self)
        else:
            return ('<{0.__class__.__name__} namespace={0.namespace}>'
                    .format(self))


class BaseItem:
    """Optional base class to implement ItemABC.

    This base class provides a common implementation for some of the
    :class:`ItemABC` interface abstract methods.  Deriving from this
    will not automatically make you an :class:`ItemABC` instance, you
    still need to either subclass or register the :class:`ItemABC` as
    well.
    """

    def __init__(self, cluster, raw):
        self._cluster = cluster
        self._raw = pyrsistent.freeze(raw)

    @property
    def cluster(self):
        """The :class:`kube.Cluster` instance this resource is bound to."""
        return self._cluster

    @property
    def raw(self):
        """The raw decoded JSON representing the resource."""
        return self._raw

    @property
    def meta(self):
        """The items' metadata as a :class:`kube.meta.ObjectMeta` instance.

        This provides access to the name, labels etc. of the resource item.
        """
        return _meta.ObjectMeta(self)

    def spec(self):
        """The spec of this resource item.

        This is the raw, decoded JSON data representing the spec of
        this resource item.
        """
        return pyrsistent.thaw(self.raw['spec'])

    def fetch(self):
        """Fetch the current version of the resource item.

        This will return a new instance of the current resource item
        at it's latest version.  This is useful to see any changes
        made to the object since it was last retrieved.

        :returns: An instance of the relevant :class:`ItemABC`
           subclass.

        :raises LookupError: If the resource item no longer exits.
        :raises kube.APIError: For errors from the k8s API server.
        """
        try:
            data = self.cluster.proxy.get(self.meta.link)
        except _error.APIError as err:
            if err.status_code == 404:
                raise LookupError from err
            else:
                raise
        return self.__class__(self.cluster, data)

    def watch(self):
        """Watch the resource item for changes.

        Only changes after the current version will be part of the
        iterator.  However it can not be guaranteed that *every*
        change is returned, if the current version is rather old some
        changes might no longer be available.

        :returns: An iterator of :class:`kube.WatchEvents` instances
           where the :attr:`kube.WatchEvent.item` attribute will be
           a :class:`kube.NodeItem` instance.
        :raises kube.APIError: For errors from the k8s API server.
        """
        return _util.watch_item(self)

    def delete(self):
        """Delete the resource item.

        If the item does not exist this will silently do nothing,
        following Kubernete's principle of being a level-based system
        and the desired state is already present so all is fine.

        :rtype: None

        :raises APIError: For errors from the k8s API server.
        """
        try:
            self.cluster.proxy.delete(self.meta.link)
        except _error.APIError as err:
            if err.status_code != 404:
                raise

    def __repr__(self):
        if self.meta.namespace is not None:
            return ('<{0.__class__.__name__} {0.meta.name} '  # pylint: disable=missing-format-attribute
                    'namespace={0.meta.namespace}>' .format(self))
        else:
            return '<{0.__class__.__name__} {0.meta.name}>'.format(self)  # pylint: disable=missing-format-attribute


class Kind(enum.Enum):
    """The kinds of Kubernetes resources."""
    Node = 'Node'
    NodeList = 'NodeList'
    Namespace = 'Namespace'
    NamespaceList = 'NamespaceList'
    Pod = 'Pod'
    PodList = 'PodList'
    ReplicaSet = 'ReplicaSet'
    ReplicaSetList = 'ReplicaSetList'
    Service = 'Service'
    ServiceList = 'ServiceList'
    Secret = 'Secret'
    SecretList = 'SecretList'



"""Interface for Secret resources."""

import base64
import enum

from kube import _base


class SecretView(_base.BaseView, _base.ViewABC):
    """View of the Secret resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.SecretList
    resource = 'secrets'


class SecretType(enum.Enum):
    """Enumeration of secret types."""
    Opaque = 'Opaque'


class SecretItem(_base.BaseItem, _base.ItemABC):
    """A Secret in the Kubernetes cluster.

    :param cluster: The cluster this Service exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    :cvar SecretType: Shortcut to :class:`kube.SecretType`.
    """
    kind = _base.Kind.Secret
    resource = 'secrets'
    SecretType = SecretType

    def spec(self):
        """An empty dictionary.

        This is supposed to be the secret resource item's spec.  But
        secrets do not have a spec, so to still follow the
        :class:`kube.ItemABC` we return an empty dict.
        """
        return {}

    @property
    def type(self):
        """The type of secret.

        There currently is only the "Opaque" type.
        """
        return self.SecretType(self.raw['type'])

    @property
    def data(self):
        """A mapping of the secret data.

        A copy of the secret data as a dict.  The keys are the names
        of the secrets as a (unicode) string, while the values are the
        secrets as bytestrings.

        Secret values are stored in a base64 encoding on the k8s
        master, but this is an implementation detail that this
        property takes care off for you.
        """
        data = {}
        for name, value in self.raw.get('data', {}).items():
            data[name] = base64.b64decode(value)
        return data



"""Interface for ReplicaSet resources.

These used to be called ReplicationController before.
"""

from kube import _base
from kube import _util


class ReplicaSetView(_base.BaseView, _base.ViewABC):
    """View of the ReplicaSet resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.ReplicaSetList
    resource = 'replicationcontrollers'


class ReplicaSetItem(_base.BaseItem, _base.ItemABC):
    """A ReplicaSet in the Kubernetes cluster.

    A ReplicaSet, formerly known as a ReplicationController, is
    responsible for keeping a desired number of pods running.

    :param cluster: The cluster this ReplicaSet exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.ReplicaSet
    resource = 'replicationcontrollers'

    @_util.statusproperty
    def observed_replicas(self):
        """The current number of replicas observed."""
        return self.raw['status']['replicas']

    @_util.statusproperty
    def observed_generation(self):
        """The (integer) generation of the ReplicaSet."""
        return self.raw['status']['observedGeneration']

    @_util.statusproperty
    def fully_labeled_replicas(self):
        """Number of pods which have an exact matching set of labels.

        This counts the pods which have the exact same set of labels
        as the labelselector of this replicaset.
        """
        return self.raw['status']['fullyLabeledReplicas']



"""Interface for ObjectMeta objects."""

import collections.abc

import pyrsistent

from kube import _util as util


class ObjectMeta:
    """Common metadata for API objects.

    :param resource: The object representing the Kubernetes resource which
       this metadata describes.
    :type resource: kube._base.ItemABC
    """

    def __init__(self, resource):
        self._resource = resource

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self._resource.raw['metadata'] ==
                    other._resource.raw['metadata'])  # pylint: disable=protected-access
        else:
            return NotImplemented

    @property
    def name(self):
        """The name of the object."""
        return self._resource.raw['metadata']['name']

    @property
    def namespace(self):
        """Namespace the object resides in, or ``None``."""
        try:
            return self._resource.raw['metadata']['namespace']
        except KeyError:
            return None

    @property
    def labels(self):
        """The labels as a :class:`ResourceLabels` instance."""
        return ResourceLabels(self._resource)

    @property
    def version(self):
        """The opaque resource version."""
        return self._resource.raw['metadata']['resourceVersion']

    @property
    def created(self):
        """The created timestamp as a :class:`datetime.datetime` instance."""
        return util.parse_rfc3339(
            self._resource.raw['metadata']['creationTimestamp'])

    @property
    def link(self):
        """A link to the resource itself.

        This is currently an absolute URL without the hostname, but
        you don't have to care about that.  The
        :class:`kube.APIServerProxy` will be just fine with it as it's
        ``path`` argument.
        """
        return self._resource.raw['metadata']['selfLink']

    @property
    def uid(self):
        """The Universal ID of the item.

        This is unique for this resource kind.
        """
        return self._resource.raw['metadata']['uid']


class ResourceLabels(collections.abc.Mapping):
    """The labels applied to an API resource item.

    This allows introspecting the labels as a normal mapping and
    provides a few methods to directly manipulate the labels on the
    resource item.
    """

    def __init__(self, resource):
        self._resource = resource
        self._data = resource.raw['metadata'].get('labels', pyrsistent.pmap())

    def __getitem__(self, name):
        return self._data[name]

    def __iter__(self):
        for key in self._data:
            yield key

    def __len__(self):
        return len(self._data)

    def set(self, key, value):
        """Set a (new) label.

        This will set or update the label's value on the resource.

        :returns: A new instance of the resource.
        :raises kube.APIError: If there is a problem with the API server.
        """
        raw = self._resource.cluster.proxy.patch(
            self._resource.meta.link,
            patch={'metadata': {'labels': {key: value}}},
        )
        return self._resource.__class__(self._resource.cluster, raw)

    def delete(self, key):
        """Delete a label.

        This will remove the label for a given key from the resource.

        :returns: A new instance of the resource.
        :raises kube.APIError: If there is a problem with the API server.
        """
        raw = self._resource.cluster.proxy.patch(
            self._resource.meta.link,
            patch={'metadata': {'labels': {key: None}}},
        )
        return self._resource.__class__(self._resource.cluster, raw)

    def __repr__(self):
        return '<Labels "{}">'.format(
            ','.join('{}={}'.format(k, v) for k, v in self.items()))



"""Interface for namespaces."""

import enum

from kube import _base
from kube import _error
from kube import _util


class NamespaceView(_base.BaseView, _base.ViewABC):
    """View of all the Namespace resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.  This is here for the :class:`kube.ViewABC`
       compatibility, namespaces can not be used for the NamespaceList
       resource.  A :class:`kube.NamespaceError` is raised when this
       is not ``None``.
    :type namespace: NoneType

    :raises kube.NamespaceError: If instantiated using a namespace.

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.NamespaceList
    resource = 'namespaces'

    def __init__(self, cluster, namespace=None):
        if namespace is not None:
            raise _error.NamespaceError(
                'NamespaceList does not support namespaces')
        super().__init__(cluster, None)

    def fetch(self, name, namespace=None):
        """Retrieve an individual Namespace resource item by name.

        This returns the current verison of the resource item.

        :param str name: The name of the namespace resource item to
           retrieve.
        :param str namespace: Must be *None* or a
           :class:`kube.NamespaceError` is raised.  Here only for
           compatibility with the ABC.

        :returns: A :class:`kube.NamespaceItem` instance.
        :raises LookupError: If the namespace does not exist.
        :raises kube.APIError: For errors from the k8s API server.
        :raises kube.NamespaceError: If a namespace is used.
        """
        if namespace is not None:
            raise _error.NamespaceError(
                'NamespaceView does not support namespaces')
        data = _util.fetch_resource(self.cluster, 'namespaces', name)
        return NamespaceItem(self.cluster, data)


class NamespacePhase(enum.Enum):
    """Enumeration of all possible namespace phases.

    This is aliased to :attr:`NamespaceResource.NamespacePhase`
    for convenience.
    """
    Active = 'Active'
    Terminating = 'Terminating'


class NamespaceItem(_base.BaseItem, _base.ItemABC):
    """A namespace in the Kubernetes cluster.

    See http://kubernetes.io/docs/admin/namespaces/ for details.

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    :cvar NamespacePhase: Convenience alias of :class:`NamespacePhase`.
    """
    kind = _base.Kind.Namespace
    resource = 'namespaces'
    NamespacePhase = NamespacePhase

    @property
    def phase(self):
        """Phase of the namespace as a :class:`kube.NamespacePhase`."""
        return NamespacePhase(self.raw['status']['phase'])



"""Interface for pod resources."""

import enum
import ipaddress

from kube import _base
from kube import _error
from kube import _util


class PodView(_base.BaseView, _base.ViewABC):
    """View of the Pod resource items in the cluster.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param namespace: Limit the view to resource items in this
       namespace.
    :type namespace: str

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    """
    kind = _base.Kind.PodList
    resource = 'pods'


class PodPhase(enum.Enum):
    """Enumeration of all possible pod phases.

    This is aliased to :attr:`Pod.PodPhase` for convenience.
    """
    Pending = 'Pending'
    Running = 'Running'
    Succeeded = 'Succeeded'
    Failed = 'Failed'
    Unknown = 'Unknown'


class PodItem(_base.BaseItem, _base.ItemABC):
    """A pod in the Kubernetes cluster.

    Each pod contains a number of containers and volumes which are executed
    on a node within the cluster. A pod may exist in a namespace. Pods are
    typically managed by a controller such as a replication controller or
    job.

    :param cluster: The cluster this pod exists in.
    :type cluster: kube.Cluster
    :param raw: The raw data of the resource item.
    :type raw: pyrsistent.PMap

    :cvar kind: The kind of the underlying Kubernetes resource item.
    :cvar resource: The name of the Kubernetes API resource.
    :cvar PodPhase: Convenience alias of :class:`PodPhase`.
    """
    kind = _base.Kind.Pod
    resource = 'pods'
    PodPhase = PodPhase

    @_util.statusproperty
    def phase(self):
        """Phase of the pod as a :class:`kube.PodPhase`.

        :raises kube.StatusError: If this status item is not present.
        """
        return PodPhase(self.raw['status']['phase'])

    @_util.statusproperty
    def start_time(self):
        """Start the pod was started as a :class:`datetime.datetime`.

        :raises kube.StatusError: If this status item is not present.
        """
        return _util.parse_rfc3339(self.raw['status']['startTime'])

    @_util.statusproperty
    def ip(self):               # pylint: disable=invalid-name
        """IP address of the pod within the cluster.

        This may be as a :class:`ipaddress.IPv4Address` or a
        :class:`ipaddress.IPV6Address`.

        :raises kube.StatusError: If this status item is not present.
        """
        return ipaddress.ip_address(self.raw['status']['podIP'])

    @_util.statusproperty
    def host_ip(self):
        """IP address of the pod's host within the cluster.

        This  may be as a :class:`ipaddress.IPv4Address` or a
        :class:`ipaddress.IPv6Address`.

        :raises kube.StatusError: If this status item is not present.
        """
        return ipaddress.ip_address(self.raw['status']['hostIP'])

    @_util.statusproperty
    def message(self):
        """Human readable message explaining the pod's state.

        :raises kube.StatusError: If this status item is not present.
        """
        return self.raw['status']['message']

    @_util.statusproperty
    def reason(self):
        """PascalCase string explaining the pod's state.

        :raises kube.StatusError: If this status item is not present.
        """
        return self.raw['status']['reason']

    @property
    def containers(self):
        """Iterate over all :class:`Container` instances in the pod."""
        statuses = self.raw['status'].get('containerStatuses', [])
        for container_status in statuses:
            yield Container(self, container_status)


class Container:
    """A container inside a pod.

    Containers live inside a pod and may be restarted inside this pod
    as controlled by the restart policy set on the pod.

    :param pod: The pod the container is part off.
    :type pod: PodItem
    :param raw: The JSON-decoded object describing the status of the
       container.
    :type raw: pyrsistent.PMap

    :ivar pod: The :class:`PodItem` instance the container is bound to.
    :ivar raw: The raw JSON-decoded object representing the container.
    """

    def __init__(self, pod, raw):
        self.pod = pod
        self.raw = raw

    def __repr__(self):
        return "<{0.__class__.__name__} {0.name!r} of {0.pod}>".format(self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.raw == other.raw
        else:
            return NotImplemented

    @property
    def name(self):
        """The name of the container as a string."""
        return self.raw['name']

    @_util.statusproperty
    def state(self):
        """Current state of the container.

        This is represented by a :class:`ContainerState` instance.

        :raises kube.StatusError: If this status item is not present.
        """
        state = ContainerState(self.raw['state'])
        return state

    @_util.statusproperty
    def last_state(self):
        """Previous state of the container, if known.

        This is represented by a :class:`ContainerState` instance.

        :raises kube.StatusError: If this status item is not present.
        """
        if 'lastState' in self.raw and self.raw['lastState']:
            return ContainerState(self.raw['lastState'])
        else:
            raise _error.StatusError('No previous container state available')

    @_util.statusproperty
    def ready(self):
        """Boolean indicating if the container passed it's readyness probe.

        :raises kube.StatusError: If this status item is not present.
        """
        return bool(self.raw['ready'])

    @_util.statusproperty
    def restart_count(self):
        """The number of times the container was restarted as an integer.

        Note that this is currently not always accurate, it counts the
        number of dead containers which have not yet been removed.
        This means the gargage collection of containers caps this
        number at 5.
        """
        return int(self.raw['restartCount'])

    @_util.statusproperty
    def image(self):
        """The image the container is running.

        For Docker this is normally the repository name with tag
        appended.
        """
        return self.raw['image']

    @_util.statusproperty
    def image_id(self):
        """The ImageID of the container's image.

        For Docker this is in the ``docker://<hex_id>`` format.
        """
        return self.raw['imageID']

    @property
    def id(self):               # pylint: disable=invalid-name
        """The ID of the running container.

        For Docker this is in the ``docker://<hex_id>`` format.
        """
        return self.raw['containerID']


class ContainerState:
    """The state of a container within a pod.

    A container can be in one of three states: *running*, *waiting* or
    *terminated*.  This class provides a uniform interface to all
    states and their associated details.  Not all fields are always
    valid for each state so they can all raise an
    :exc:`kube.StatusError` when they are not available or not
    applicable.

    The overall state of the container is available both as a string
    in the :attr:`state` attribute as well as booleans in the
    :attr:`waiting`, :attr:`running` and :attr:`terminated`
    attributes.

    :param raw: The raw JSON-decoded ``v1.ContainerState`` API object
        as exposed by ``v1.ContainerStatus`` objects.
    :type raw: pyrsistent.PMap

    :ivar raw: The raw JSON-decoded object representing the container state.
    """

    def __init__(self, raw):
        # Sometimes k8s gets it wrong and ends up with multiple states
        # instead of just the one.  In this case the state in
        # lastState will still be present in the existing state.  So
        # we prefer states in a certain order.
        states = set(raw.keys())
        if len(states) > 1:
            if states == {'waiting', 'terminated'}:
                self._state = 'waiting'
            elif states == {'waiting', 'running'}:
                self._state = 'running'
            elif states == {'running', 'terminated'}:
                self._state = 'terminated'
            else:
                raise ValueError('Unknown state combination')
        else:
            self._state = states.pop()
        self._data = raw[self._state]
        self.raw = raw

    def __repr__(self):
        return '<{0.__class__.__name__} {0._state}>'.format(self)

    @property
    def waiting(self):
        """Boolean indicating if the container is waiting."""
        return self._state == 'waiting'

    @property
    def running(self):
        """Boolean indicating if the container is running."""
        return self._state == 'running'

    @property
    def terminated(self):
        """Boolean indicating if the container has been terminated."""
        return self._state == 'terminated'

    @_util.statusproperty
    def reason(self):
        """Brief reason explaining the container's state (str).

        This is normally a CamelCased message ID.

        Available for *waiting* and *terminated* states.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['reason']

    @_util.statusproperty
    def message(self):
        """Message regarding the container's state (str).

        Available for *waiting* and *terminated* states.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['message']

    @_util.statusproperty
    def started_at(self):
        """The time the container was started or restarted (datetime.datetime).

        Available for the *running* state.

        :raises kube.StatusError: When this is not provided.
        """
        return _util.parse_rfc3339(self._data['startedAt'])

    @_util.statusproperty
    def exit_code(self):
        """Exit code of the container (int).

        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['exitCode']

    @_util.statusproperty
    def signal(self):
        """Last signal sent to the container, if known (int).

        Not all terminated containers can be expected to have this.

        .. warning::
            The signal is identified numerically, however these signal
            numbers are not portable therefore it's ill-advised to attempt
            to compare this value with the constants provided by the
            built-in :mod:`singal` module.

        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['signal']

    @_util.statusproperty
    def finished_at(self):
        """The time the container was terminated (datetime.datetime).

        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return _util.parse_rfc3339(self._data['finishedAt'])

    @_util.statusproperty
    def container_id(self):
        """The container ID of the terminated container.
        Available for the *terminated* state.

        :raises kube.StatusError: When this is not provided.
        """
        return self._data['containerID']



"""Common utilities kube."""

import collections
import datetime
import functools
import json

import pyrsistent

from kube import _error
from kube import _watch


def parse_rfc3339(when):
    """Parse a RFC 3339 timestamp into a :class:`datetime.datetime`."""
    return datetime.datetime.strptime(when, '%Y-%m-%dT%H:%M:%SZ')


def fetch_resource(cluster, resource, name, namespace=None):
    """Return the single resource data.

    This method helps implement common functionality for the
    :meth:`kube.ViewABC.fetch` method.  In particular it handles 404 as
    aa :class:`LookupError`.

    .. warning::

       Take care not to use this without a namespace when the resource
       does not support this, as this will result in a
       :class:`KeyError` instead of :class:`kube.APIError`.

    :returns: The decoded JSON object of the resource.
    :raises LookupError: if the resource does not exist.
    :raises kube.APIError: for errors from the k8s API server.
    """
    try:
        if namespace:
            data = cluster.proxy.get('namespaces', namespace, resource, name)
        else:
            data = cluster.proxy.get(resource, name)
    except _error.APIError as exc:
        if exc.status_code == 404:
            raise LookupError('No such resource: {}'.format(resource)) from exc
        else:
            raise
    else:
        return data


def filter_list(cluster, resource, labels, fields, namespace=None):
    """Return an iterator for a resource restricted to the selector.

    This method helps implement common functionality for the
    :meth:`kube.ViewABC.filter` method.

    :param cluster: The cluster instance.
    :type cluster: kube.Cluster
    :param resource: The resource name, defining the API server
       endpoint to use.
    :type resource: str
    :param labels: A label selector expression.  This can either
       be a dictionary with labels which must match exactly, or a
       string label expression as understood by k8s itself.
    :type labels: dict or str
    :param fields: A field selector expression.  This can either
       be a dictionary with fields which must match exactly, or a
       string field selector as understood by k8s itself.
    :type fields: dict or str
    :param namespace: Only include items in the given namespace.
    :type namespace: str

    :returns: an iterator of decoded JSON objects of resources
       matching the selector.
    :rtype: generator

    :raises ValueError: if an empty selector is used.
    :raises kube.APIError: for errors from the k8s API server.
    """
    if isinstance(labels, collections.abc.Mapping):
        if len(labels) == 0:
            raise ValueError('Empty labels selector')
        labels = ','.join('{}={}'.format(k, v) for k, v in labels.items())
    if isinstance(fields, collections.abc.Mapping):
        if len(fields) == 0:
            raise ValueError('Empty field selector')
        fields = ','.join('{}={}'.format(k, v) for k, v in fields.items())
    params = {}
    if labels:
        params['labelSelector'] = labels
    if fields:
        params['fieldSelector'] = fields
    if namespace:
        data = cluster.proxy.get('namespaces', namespace, resource, **params)
    else:
        data = cluster.proxy.get(resource, **params)
    yield from data['items']


def watch_item(item):
    """Watch a resource item for changes.

    Only changes after the current version will be part of the
    iterator.  However it can not be guaranteed that *every* change is
    returned, if the current version is rather old some changes might
    no longer be available.

    :param item: The Resource Item to watch.
    :type item: kube.ItemABC

    :returns: An iterator of the :class:`kube.WatchEvent` instances
       where the :attr:`kube.WatchEvent.item` attribute will be an
       instance of the class of the ``item`` param.

    :raises kube.APIError: For errors from the k8s API server.
    """
    fieldsel = {'metadata.uid': item.meta.uid}
    listurl = '/'.join(item.meta.link.split('/')[:-1])
    jsonwatcher = item.cluster.proxy.watch(listurl, fields=fieldsel)
    jsonwatcher.skip(item.meta.version)
    return _watch.ResourceWatcher(item.cluster, jsonwatcher, item.__class__)


def freeze(obj):
    """Freeze an object.

    This tries to make an object consisting of dicts, lists and
    primitive types immutable using pyrsistent.
    """
    return pyrsistent.freeze(obj)


def thaw(obj):
    """Thaw an objct.

    The reverse of :func:`freeze`.
    """
    return pyrsistent.thaw(obj)


def statusproperty(func):
    """Easily create properties for k8s status items.

    This turns the function into a property and will catch any
    exceptions, wrapping them into a :class:`kube.StatusError`.  This
    is far from perfect as it could easily be masking real bugs as
    well, the :attr:`StatusError.__cause__` attribute can still be
    used to find the original exception.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):  # pylint: disable=missing-docstring
        try:
            return func(*args, **kwargs)
        except Exception as err:
            raise _error.StatusError from err
    return property(wrapper)


class ImmutableJSONDecoder(json.JSONDecoder):
    """Decode JSON to immutable types."""

    def decode(self, string, *args, **kwargs):
        return freeze(super().decode(string, *args, **kwargs))
